//==============================================================================================
//
//      This file is part of LiDIA --- a library for computational number theory
//
//      Copyright (c) 1994--2001 the LiDIA Group.  All rights reserved.
//
//      See http://www.informatik.tu-darmstadt.de/TI/LiDIA/
//
//----------------------------------------------------------------------------------------------
//
//      $Id$
//
//      Author  : Volker Mueller (VM), based on an implementation
//                of Thomas Sosnowski
//      Changes : See CVS log
//
//==============================================================================================



#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "LiDIA/LiDIA.h"

#ifdef LIDIA_NAMESPACE
namespace LiDIA {
#endif

  // defined in base/factorization/bigint/sf_mpqs.cc
  extern double const mpqs_machine_factor; 

#ifdef LIDIA_NAMESPACE
} // namespace LiDIA 
#endif


